from typing import List, Tuple, NamedTuple


class Bound(NamedTuple):
    version: str
    inclusive: bool = False

# FIXME: change implemntation of Version - str
class Version(object):
    def __init__(self, version):
        self.version = version

    def __str__(self):
        return f"{self.version}"

    def __eq__(self, other):
        if isinstance(other, Version):
            return self.version == other.version
        else:
            return False


class Restriction(object):

    @classmethod
    def all_versions(cls):
        return cls(Version(None), False, Version(None), False)

    # FIXME: change constructor to use `Bound`
    def __init__(self, lower_bound: Version, has_inclusive_lower: bool,
                 upper_bound: Version, has_inclusive_upper: bool):
        if not all((isinstance(lower_bound, Version),
                    isinstance(upper_bound, Version))):
            raise ValueError(
                "lower_bound and upper_bound must be of type Version")

        self.lower_bound = lower_bound
        self.has_inclusive_lower = has_inclusive_lower
        self.upper_bound = upper_bound
        self.has_inclusive_upper = has_inclusive_upper

    def __str__(self):
        buffer = ['[' if self.has_inclusive_lower else '(']
        if self.lower_bound.version:
            buffer.append(str(self.lower_bound))
        buffer.append(',')
        if self.upper_bound.version:
            buffer.append(str(self.upper_bound))
        buffer.append(']' if self.has_inclusive_upper else ')')
        # check if the versions are equals - [X.X.X,X.X.X] -> [X.X.X]
        if len(buffer) == 5 and buffer[1] == buffer[3]:
            buffer = ['[', str(self.lower_bound), ']']
        return "".join(buffer)

    def __eq__(self, other):
        if isinstance(other, Restriction):
            lower_eq = (self.lower_bound == other.lower_bound)
            upper_eq = (self.upper_bound == other.upper_bound)
            inclusive_lower_eq = (
                    self.has_inclusive_lower == other.has_inclusive_lower)
            inclusive_upper_eq = (
                    self.has_inclusive_upper == other.has_inclusive_upper)
            return lower_eq and upper_eq and inclusive_lower_eq and inclusive_upper_eq
        else:
            # fixme: raise exception
            return False

    @property
    def bounds(self) -> Tuple[Bound, Bound]:
        lower = Bound(
            version=self.lower_bound.version,
            inclusive=self.has_inclusive_lower
        )
        upper = Bound(
            version=self.upper_bound.version,
            inclusive=self.has_inclusive_upper
        )
        # FIXME: maybe handle equal lower/upper `[VER]`
        return lower, upper


class UnifiedVersionRange(object):
    # FIXME: REMOVE recommended_version feature - unused
    def __init__(self, recommended_version, restrictions):
        self.recommended_version = recommended_version
        self.restrictions = restrictions

    def __str__(self):
        if self.recommended_version is not None:
            return str(self.recommended_version)
        else:
            buffer = []
            for r in self.restrictions:
                buffer.append(str(r))
            return ",".join(buffer)

    def __eq__(self, other):
        restrictions_eq = self.restrictions == other.restrictions
        return restrictions_eq

    @property
    def constraints(self) -> List[Restriction]:
        return self.restrictions

    @staticmethod
    def parse_restriction(spec):
        has_inclusive_lower = spec.startswith("[")
        has_inclusive_upper = spec.endswith("]")
        process = spec[1:-1].strip()
        index = process.find(',')
        # [X.X.X]
        if index < 0:
            if not has_inclusive_lower or not has_inclusive_upper:
                raise ValueError(
                    "Single version must be surrounded by []: {}".format(spec))
            version = Version(process)
            restriction = Restriction(version, has_inclusive_lower, version,
                                      has_inclusive_upper)
        else:
            lower_bound = process[0: index].strip()
            upper_bound = process[index + 1:].strip()
            if lower_bound == '' and upper_bound == '':
                return Restriction.all_versions()
            if lower_bound == upper_bound:
                raise ValueError(
                    "Range cannot have identical boundaries: {}".format(spec))
            lower_version = Version(None)
            if len(lower_bound) > 0:
                lower_version = Version(lower_bound)
            upper_version = Version(None)
            if len(upper_bound) > 0:
                upper_version = Version(upper_bound)
            restriction = Restriction(lower_version, has_inclusive_lower,
                                      upper_version, has_inclusive_upper)
        return restriction

    @staticmethod
    def create_from_spec(spec: str):
        """
        Create unifiedVersionRange from maven spec string - `[x,y]`
        :param spec:
        :return:
        """
        restrictions = []
        process = spec.strip()
        version = None
        if not spec:
            return
        if not isinstance(spec, str):
            raise ValueError(
                "Only Strings allowed.\ninput spec:{}\ntype: {}".format(
                    spec,
                    type(spec)
                )
            )

        if not process.startswith(("(", "[")) and not process.endswith(
                (")", "]")):
            raise ValueError("Recommended Version is currently not supported.")
        while process.startswith("[") or process.startswith("("):
            index1 = process.find(")")
            index2 = process.find("]")
            index = index2
            if index2 < 0 or index1 < index2:
                if index1 >= 0:
                    index = index1
            if index < 0:
                raise ValueError("Unbounded range: {}".format(spec))

            restriction = UnifiedVersionRange.parse_restriction(
                process[0:index + 1])
            restrictions.append(restriction)
            process = process[index + 1:].strip()
            if len(process) > 0 and process.startswith(","):
                process = process[1:].strip()
        if len(process) > 0:
            if len(restrictions) > 0:
                raise ValueError(
                    "Only fully-qualified sets allowed in multiple set scenario: {}".format(
                        spec))
            else:
                # fixme: use strings instead Version
                version = Version(process)
                restrictions.append(Restriction.all_versions())

        return UnifiedVersionRange(version, restrictions)

    # FIXME: Remove not needed
    @staticmethod
    def create_from_version(version):
        restrictions = []
        if not isinstance(version, str):
            raise ValueError(
                "Only Strings allowed.\ninput version:{}\ntype: {}".format(
                    version, type(version)))
        return UnifiedVersionRange(Version(version), restrictions)



from typing import List

from unified_range import utils

from unified_range.models import UnifiedVersionRange
from unified_range.utils import is_semver_range, is_unified_range


def from_semver(semver_spec: str) -> str:
    """
    Convert semver string to unified range string.
    :param semver_spec: str
    :return: unified_spec
    """
    ver_rng = utils.create_from_semver(semver_spec)
    return str(ver_rng)


def to_semver(spec: str) -> str:
    """
    Convert unified range string to semver string.
    :param spec: str
    :return: semver
    """
    semver = utils.transform_to_semver(spec, separator=" ")
    return semver


def to_semver_comma_separated(spec: str) -> str:
    """
    Convert unified range string to semver string, with
    comma (and space) separating restrictions in case there are
    2. For example: `>1.2.3, <=4.5.6`
    :param spec: str
    :return: semver
    """
    semver = utils.transform_to_semver(spec, separator=", ")
    return semver


def unified_range(spec: str) -> UnifiedVersionRange:
    """
    Return VersionRange for unified range.
    Only support unified range format.
    :param spec: str
    :return: VersionRange instance
    """
    return UnifiedVersionRange.create_from_spec(spec)


def filter_versions(asc_versions: List[str], ranges: List[str]) -> List[str]:
    """
    Return an ordered list of versions that not satisfies any range.
    Input versions must be ordered in ascending order and include all
    the versions that are specified in the ranges.
    :param ord_versions:
    :param ranges:
    :return:
    """
    rngs_unified = []
    for rng in ranges:
        if is_semver_range(rng):
            rngs_unified.append(unified_range(from_semver(rng)))
        elif is_unified_range(rng):
            rngs_unified.append(unified_range(rng))
        else:
            raise ValueError(
                f'Not a valid semver or unified/maven range - ({rng})')

    return utils.not_included_versions(asc_versions, rngs_unified)


def next_filtered_version(current_version: str, asc_versions: List[str],
                          ranges: List[str]) -> List[str]:
    """
    Return the first version that not satisfies any range.
    Input versions must be ordered in ascending order and include all
    the versions that are specified in the ranges.
    """
    if current_version not in asc_versions:
        raise ValueError('current_version given is not part of asc_version')
    minimal_version = None
    index_current_version = asc_versions.index(current_version)
    filtered_versions = filter_versions(asc_versions, ranges)
    for v in filtered_versions:
        if index_current_version <= asc_versions.index(v):
            minimal_version = v
            break
    return minimal_version


def maximum_filtered_version(asc_versions: List[str],
                             ranges: List[str]) -> List[str]:
    """
    Return the first version that not satisfies any range.
    Input versions must be ordered in ascending order and include all
    the versions that are specified in the ranges.
    """
    filtered_versions = filter_versions(asc_versions, ranges)
    if filtered_versions:
        return filtered_versions[-1]
    else:
        return None



import re
from typing import List, Optional

from unified_range.models import UnifiedVersionRange, Restriction, Version

semver_operators = {"lt": "<", "lte": "<=", "gt": ">", "gte": ">=", "eq": "="}
unified_operators = {"lt": ")", "lte": "]", "gt": "(", "gte": "["}


def _is_unified_ops(rng):
    return any(op in rng for op in unified_operators.values())


def _is_semver_ops(rng):
    return any(op in rng for op in semver_operators.values())


def _clean_semver(semver):
    # also cleaning `v=X.X.X` `= X.X.X`
    # TODO: regex catching invalid semvers `<=v1.2.3` => `<1.2.3'
    remove_eq = re.sub(r'(?<![<>])=?v?(?=.*?)', '', semver.strip())
    remove_eq_v = re.sub(r'(?<=[<>])v?(?=.*?)', '', remove_eq)
    return remove_eq_v


def _comparator_trim(semver: str) -> str:
    """
    clean spaces between operator to the version.
    `> 1.2.3 <= 1.2.5` => `>1.2.3 <=1.2.5`
    :param semver:
    :return:
    """
    # switch `,` to ` ` for composer versions
    if ',' in semver:
        semver = semver.replace(',', ' ')
    COMPARTOR_TRIM = r'\s+(?=[^<>|])'
    return re.sub(COMPARTOR_TRIM, '', semver.strip())


def is_semver_range(rng):
    return _is_semver_ops(rng) and not _is_unified_ops(rng)


def is_unified_range(rng):
    return _is_unified_ops(rng) and not _is_semver_ops(rng)


def transform_to_semver(unified_spec: str, separator: str) -> str:
    """
    Transform unified spec (following the maven VersioRange spec,
    https://maven.apache.org/enforcer/enforcer-rules/versionRanges.html)
    to semver range string representation (following npm/node spec,
    https://github.com/npm/node-semver#ranges).

    (1.2.3, 2.4.6] ----> >1.2.3 <=2.4.6
    Or (when separator = ', ')
    (1.2.3, 2.4.6] ----> >1.2.3, <=2.4.6

    :param unified_spec: str
    :param separator: str
    :return: semver: string
    """
    # FIXME: use semver_operators and _is_semver_{ops,range}
    operators = {"lt": "<", "lte": "<=", "gt": ">", "gte": ">="}

    if is_semver_range(unified_spec):
        raise ValueError(
            "Version ranges seems to already be semver")
    contains_all_version = UnifiedVersionRange(None, [Restriction.all_versions()])
    unified_version = UnifiedVersionRange.create_from_spec(unified_spec)
    if unified_version == contains_all_version:
        return "*"

    unified_restrictions = unified_version.restrictions
    semvers = []

    for restriction in unified_restrictions:
        if restriction.upper_bound.version and restriction.lower_bound.version:
            # specific version
            if restriction.lower_bound.version == restriction.upper_bound.version:
                semvers.append("{}".format(restriction.upper_bound))
            # two constraints semver `>1.1.2 <=2.0.0`
            # `{operator}{version} {operator}{version}`
            else:
                gt_gte = operators["gt"]
                if restriction.has_inclusive_lower:
                    gt_gte = operators["gte"]
                lt_lte = operators["lt"]
                if restriction.has_inclusive_upper:
                    lt_lte = operators["lte"]
                semvers.append(
                    "{}{}{}{}{}".format(gt_gte, restriction.lower_bound,
                                        separator,
                                        lt_lte, restriction.upper_bound))
        # one constraint semver `>=1.2.3`
        # {operator}{version}
        else:
            if restriction.upper_bound.version and not restriction.lower_bound.version:
                lt_lte = operators["lt"]
                if restriction.has_inclusive_upper:
                    lt_lte = operators["lte"]
                semvers.append(
                    "{}{}".format(lt_lte, restriction.upper_bound))
            elif restriction.lower_bound.version and not restriction.upper_bound.version:
                gt_gte = operators["gt"]
                if restriction.has_inclusive_lower:
                    gt_gte = operators["gte"]
                semvers.append(
                    "{}{}".format(gt_gte, restriction.lower_bound))
            else:
                raise ValueError("lower and upper bound are None")
    return " || ".join(semvers)


def create_from_semver(semver: str) -> UnifiedVersionRange:
    """
    Transform semver range string (following npm/node spec,
    https://github.com/npm/node-semver#ranges).

    to unified VersionRange (following the maven VersioRange spec,
    https://maven.apache.org/enforcer/enforcer-rules/versionRanges.html).

    >1.2.3 <=2.4.6 ----> (1.2.3, 2.4.6]
    :param semver: str
    :return: VersionRange
    """
    operators = {"lt": "<", "lte": "<=", "gt": ">", "gte": ">="}

    if is_unified_range(semver):
        raise ValueError(
            "Version ranges seems to already be maven version range")
    if not any(op in semver for op in operators.values()):
        semver = _clean_semver(semver)
    else:
        semver = _comparator_trim(semver)

    if semver == "*":
        return UnifiedVersionRange(None, [Restriction.all_versions()])
    semver_restrictions = semver.split("||")

    restrictions = []
    for semver_restriction in semver_restrictions:
        constraints = semver_restriction.split()
        lower_bound = None
        has_inclusive_lower = False
        upper_bound = None
        has_inclusive_upper = False
        for constraint in constraints:
            if constraint.count("<") > 1 or constraint.count(">") > 1:
                raise ValueError("semver range contains </> more than one time.")
            if constraint.startswith(operators["lte"]):
                upper_bound = constraint[2:]
                has_inclusive_upper = True
            elif constraint.startswith(operators["gte"]):
                lower_bound = constraint[2:]
                has_inclusive_lower = True
            elif constraint.startswith(operators["lt"]):
                upper_bound = constraint[1:]

            elif constraint.startswith(operators["gt"]):
                lower_bound = constraint[1:]
            else:
                upper_bound = constraint
                lower_bound = constraint
                has_inclusive_lower = True
                has_inclusive_upper = True

        lower_version = Version(None)
        upper_version = Version(None)
        if lower_bound:
            lower_version = Version(lower_bound)
        if upper_bound:
            upper_version = Version(upper_bound)
        restrictions.append(
            Restriction(lower_version, has_inclusive_lower, upper_version, has_inclusive_upper))

    return UnifiedVersionRange(None, restrictions)


def not_included_versions(ordered_version_list: List[str],
                          ranges_list: List[UnifiedVersionRange]) -> List[str]:
    """
    Filter versions that are not included in the ranges.
    Versions list must be ordered to filter correctly.
    :param ordered_version_list:
    :param ranges_list:
    :return:
    """

    def _get_index(ver_lst: List[str], ver: Optional[str],
                   include: bool = False) -> int:
        """
        get index of item in list
        """
        if ver is None:
            return

        if ver in ver_lst:
            # only first one that found
            ret = ver_lst.index(ver)
        else:
            raise ValueError(
                f"Version {ver} couldn't be found in the versions list {ver_lst}")

        if include:
            ret = ret + 1
        return ret

    # FIXME: Can be set?
    rst_indices: List[set] = []
    last_index = len(ordered_version_list)
    first_index = 0
    # FIXME: remove operators
    for rng in ranges_list:
        # calculate index of the ordered_version_list slices
        # using new property `constraints` isn't necessary
        for rst in rng.constraints:
            lower, upper = rst.bounds
            if lower == upper:
                if not lower.version and not upper.version:
                    # (,) [,] - all version included
                    return []
                # Exact version range - `[VER]`
                # lower or upper versions can be taken, and inclusive set to false
                # to get the right index.
                exact_index = _get_index(ordered_version_list, lower.version)
                # exact index to be removed is added as a set to rst_indices
                rst_indices.append(set((exact_index,)))
                continue
            lower_index = _get_index(ordered_version_list, lower.version,
                                     not lower.inclusive)
            upper_index = _get_index(ordered_version_list, upper.version,
                                     upper.inclusive)
            # use of on the list and intersection .. [:]
            # [X,Y]
            if lower_index is not None and upper_index is not None:
                rst_indices.append(set(range(lower_index, upper_index)))
            # [X,]
            elif lower_index is not None and upper_index is None:
                rst_indices.append(set(range(lower_index, last_index)))
            # [,Y]
            elif lower_index is None and upper_index is not None:
                rst_indices.append(set(range(first_index, upper_index)))

    if rst_indices:
        ind_to_remove = set.union(*rst_indices)
    else:
        ind_to_remove = []

    not_included = [v for i, v in enumerate(ordered_version_list) if
                    i not in ind_to_remove]
    return not_included



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['unified_range']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'unified-range',
    'version': '0.1.1',
    'description': 'Convert between semver range and maven version range',
    'long_description': "# unified-range\n\nThe library converts input semver ranges to a uniform model, and the other way around, providing objects that are easier to use programmatically.\n\n## Examples of supported ranges\n1. npm style semver - `<1.2.3 >=2.0.0`\n2. ruby style semver - `<1.2.3, >=2.0.0`\n3. maven style version ranges - `[1.2.3,2.1.1), [3.0.0,4.1.1)`\n\n\nAdditionally, use this library to run algorithms on any input version ranges and calculate whether a specific version is included in this range.\n\n\n## Prerequisites\n\n1. Ensure you have installed either pip or pipenv\n2. Install:\n   `pipenv install unified-range` or `pip install unified-range`\n\n3. Import the `api` module:\n   `from unified_range import api`\n\n## How to use\nFollowing are the different functions you can perform with this library.\n\n\n### To convert a range to the uniform string range, from the semver format:\n\n`ver_rng = api.from_semver(semver_str)`\n\nResults: uniform range structure\n\n### Convert from the uniform range structure to a semver string (return str):\n\n`semver = api.to_semver(unified_spec_str)`\n\n\n### To convert the versionrange object to a string:\n\n`version_range_str = str(ver_rng)`\n\n\n### Convert from the uniform string to the uniform model object (VersionRange objects):\n\n`ver_rng = api.unified_range(unified_spec_str)`\n\n```\n>>> api.unified_range('[1.2.3,4.5.6)')\n<unified_range.models.UnifiedVersionRange at 0x7f7e4dc17320>\n```\n\n### Within a list of ranges, retrieve versions not included:\n\n`filtered_lst = api.filter_versions(ascending_version_list, ranges)`\n```\n>>> api.filter_versions(['0.1', '0.2', '1.0', '1.1', '2.0'], ['[,0.2]', '[1.1]'])\n['1.0', '2.0']\n```\n\n\nThe versions in `ascending_version_list` should be sorted in ascending order,\nfrom oldest to newest, and contain all the versions for the package.\n\n\n### From a list of version ranges, retrieve the closest version in the list to the current version (next):\nFilter next version and maximum version from list of version and ranges:\n\n`next_version = api.next_filtered_version(current_version, ascending_version_list, ranges)`\ncurrent_version must be included in the ascending_version_list.\n```\n>>> api.next_filtered_version(current_version='0.2', ascending_version_list=['0.1', '0.2', '1.0', '1.1', '2.0'], ranges=['[,0.2]', '[1.1]'])\n'1.0'\n\n>>> api.next_filtered_version(current_version='1.1', ascending_version_list=['0.1', '0.2', '1.0', '1.1', '2.0'], ranges=['[,0.2]', '[1.1]'])\n'2.0'\n ```\n\n### Retreive the latest version that is not included:\n`max_version = api.maximum_filtered_version(ascending_version_list, ranges)`\n```\n>>> api.maximum_filtered_version(ascending_version_list=['0.1', '0.2', '1.0', '1.1', '2.0'], ranges=['[,0.2]', '[1.1]'])\n'2.0'\n ```\n\n## Uniform structure examples\n\nFollowing are the uniform structures used in this library:\n\nUniform string structure example: (,1.2.3)\n\n### Uniform model examples:\n\n`UnifiedVersionRange.constraints -> List[Restrictions]`\n\n`Restriction.bounds -> Tuple[Bound, Bound]`\n\n`Bound.version -> str`\n\n`Bound.inclusive -> boolean`\n\n\n## References and prior works\n\nThis library was built with the following:\n1. Maven’s VersionRange:\n[model](https://github.com/apache/maven/tree/master/maven-artifact/src/main/java/org/apache/maven/artifact/versioning) and [spec](https://maven.apache.org/enforcer/enforcer-rules/versionRanges.html) of maven.\n2. https://semver.org/\n3. [npm’s semver library](https://www.npmjs.com/package/semver )\n",
    'author': 'Snyk',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/snyk/unified-range',
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
